import base64
from dataclasses import dataclass
import io
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from urllib.parse import urlparse

import PIL
import requests
from PIL.Image import Image as PILImage


class ControlTokenOverlap(Enum):
    """
    What to do if a control partially overlaps with a text or image token.

    Partial:
        The factor will be adjusted proportionally with the amount of the token
        it overlaps. So a factor of 2.0 of a control that only covers 2 of 4
        token characters, would be adjusted to 1.5.

    Complete:
        The full factor will be applied as long as the control overlaps with
        the token at all. How many explanations should be returned in the output.
    """

    Partial = "partial"
    Complete = "complete"

    def to_json(self) -> str:
        return self.value


@dataclass(frozen=True)
class TokenControl:
    """
    Used for Attention Manipulation, for a given token index, you can supply
    the factor you want to adjust the attention by.

    Parameters:
        pos (int, required):
            The index of the token in the prompt item that you want to apply
            the factor to.

        factor (float, required):
            The amount to adjust model attention by.
            Values between 0 and 1 will suppress attention.
            A value of 1 will have no effect.
            Values above 1 will increase attention.

    Examples:
        >>> Tokens([1, 2, 3], controls=[TokenControl(pos=1, factor=0.5)])
    """

    pos: int
    factor: float

    def to_json(self) -> Mapping[str, Any]:
        return {"index": self.pos, "factor": self.factor}


@dataclass(frozen=True)
class Tokens:
    """
    A list of token ids to be sent as part of a prompt.

    Parameters:
        tokens (List(int), required):
            The tokens you want to be passed to the model as part of your prompt.

        controls (List(TokenControl), optional, default None):
            Used for Attention Manipulation. Provides the ability to change
            attention for given token ids.

    Examples:
        >>> token_ids = Tokens([1, 2, 3], controls=[])
        >>> prompt = Prompt([token_ids])
    """

    tokens: Sequence[int]
    controls: Sequence[TokenControl]

    def to_json(self) -> Mapping[str, Any]:
        """
        Serialize the prompt item to JSON for sending to the API.
        """
        return {
            "type": "token_ids",
            "data": self.tokens,
            "controls": [c.to_json() for c in self.controls],
        }

    @staticmethod
    def from_json(json: Mapping[str, Any]) -> "Tokens":
        return Tokens(tokens=json["data"], controls=[])

    @staticmethod
    def from_token_ids(token_ids: Sequence[int]) -> "Tokens":
        return Tokens(token_ids, [])


@dataclass(frozen=True)
class TextControl:
    """
    Attention manipulation for a Text PromptItem.

    Parameters:
        start (int, required):
            Starting character index to apply the factor to.
        length (int, required):
            The amount of characters to apply the factor to.
        factor (float, required):
            The amount to adjust model attention by.
            Values between 0 and 1 will suppress attention.
            A value of 1 will have no effect.
            Values above 1 will increase attention.
        token_overlap (ControlTokenOverlap, optional):
            What to do if a control partially overlaps with a text token.

            If set to "partial", the factor will be adjusted proportionally
            with the amount of the token it overlaps. So a factor of 2.0 of a
            control that only covers 2 of 4 token characters, would be adjusted
            to 1.5.

            If set to "complete", the full factor will be applied as long as
            the control overlaps with the token at all.

            If not set, the API will default to "partial".
    """

    start: int
    length: int
    factor: float
    token_overlap: Optional[ControlTokenOverlap] = None

    def to_json(self) -> Mapping[str, Any]:
        payload: Dict[str, Any] = {
            "start": self.start,
            "length": self.length,
            "factor": self.factor,
        }
        if self.token_overlap is not None:
            payload["token_overlap"] = self.token_overlap.to_json()
        return payload


@dataclass(frozen=True)
class Text:
    """
    A Text-prompt including optional controls for attention manipulation.

    Parameters:
        text (str, required):
            The text prompt
        controls (list of TextControl, required):
            A list of TextControls to manipulate attention when processing the prompt.
            Can be empty if no manipulation is required.

    Examples:
        >>> Text("Hello, World!", controls=[TextControl(start=0, length=5, factor=0.5)])
    """

    text: str
    controls: Sequence[TextControl]

    def to_json(self) -> Mapping[str, Any]:
        return {
            "type": "text",
            "data": self.text,
            "controls": [control.to_json() for control in self.controls],
        }

    @staticmethod
    def from_json(json: Mapping[str, Any]) -> "Text":
        return Text.from_text(json["data"])

    @staticmethod
    def from_text(text: str) -> "Text":
        return Text(text, [])


@dataclass(frozen=True)
class Cropping:
    """
    Describes a quadratic crop of the file.
    """

    upper_left_x: int
    upper_left_y: int
    size: int


@dataclass(frozen=True)
class ImageControl:
    """
    Attention manipulation for an Image PromptItem.

    All coordinates of the bounding box are logical coordinates (between 0 and 1) and relative to
    the entire image.

    Keep in mind, non-square images are center-cropped by default before going to the model. (You
    can specify a custom cropping if you want.). Since control coordinates are relative to the
    entire image, all or a portion of your control may be outside the "model visible area".

    Parameters:
        left (float, required):
            x-coordinate of top left corner of the control bounding box.
            Must be a value between 0 and 1, where 0 is the left corner and 1 is the right corner.
        top (float, required):
            y-coordinate of top left corner of the control bounding box
            Must be a value between 0 and 1, where 0 is the top pixel row and 1 is the bottom row.
        width (float, required):
            width of the control bounding box
            Must be a value between 0 and 1, where 1 means the full width of the image.
        height (float, required):
            height of the control bounding box
            Must be a value between 0 and 1, where 1 means the full height of the image.
        factor (float, required):
            The amount to adjust model attention by.
            Values between 0 and 1 will suppress attention.
            A value of 1 will have no effect.
            Values above 1 will increase attention.
        token_overlap (ControlTokenOverlap, optional):
            What to do if a control partially overlaps with an image token.

            If set to "partial", the factor will be adjusted proportionally
            with the amount of the token it overlaps. So a factor of 2.0 of a
            control that only half of the image "tile", would be adjusted to
            1.5.

            If set to "complete", the full factor will be applied as long as
            the control overlaps with the token at all.

            If not set, the API will default to "partial".
    """

    left: float
    top: float
    width: float
    height: float
    factor: float
    token_overlap: Optional[ControlTokenOverlap] = None

    def to_json(self) -> Mapping[str, Any]:
        payload = {
            "rect": {
                "left": self.left,
                "top": self.top,
                "width": self.width,
                "height": self.height,
            },
            "factor": self.factor,
        }
        if self.token_overlap is not None:
            payload["token_overlap"] = self.token_overlap.to_json()
        return payload


@dataclass(frozen=True)
class Image:
    """
    An image send as part of a prompt to a model. The image is represented as
    base64.

    Note: The models operate on square images. All non-square images are center-cropped
    before going to the model, so portions of the image may not be visible.

    You can supply specific cropping parameters if you like, to choose a different area
    of the image than a center-crop. Or, you can always transform the image yourself to
    a square before sending it.

    Examples:
        >>> # You need to choose a model with multimodal capabilities for this example.
        >>> url = "https://cdn-images-1.medium.com/max/1200/1*HunNdlTmoPj8EKpl-jqvBA.png"
        >>> image = Image.from_url(url)
    """

    # We use a base_64 representation, because we want to embed the image
    # into a prompt send in JSON.
    base_64: str
    cropping: Optional[Cropping]
    controls: Sequence[ImageControl]

    @classmethod
    def from_image_source(
        cls,
        image_source: Union[str, Path, bytes],
        controls: Optional[Sequence[ImageControl]] = None,
    ):
        """
        Abstraction on top of the existing methods of image initialization.
        If you are not sure what the exact type of your image, but you know it is either a Path object, URL, a file path,
        or a bytes array, just use the method and we will figure out which of the methods of image initialization to use
        """
        if isinstance(image_source, Path):
            return cls.from_file(path=str(image_source), controls=controls)

        elif isinstance(image_source, str):
            try:
                p = urlparse(image_source)
                if p.scheme:
                    return cls.from_url(url=image_source, controls=controls)
            except Exception:
                # we assume that If the string runs into a Exception it isn't not a valid ulr
                pass

            return cls.from_file(path=image_source, controls=controls)

        elif isinstance(image_source, bytes):
            return cls.from_bytes(bytes=image_source, controls=controls)

        else:
            raise TypeError(
                f"The image source: {image_source} should be either Path, str or bytes"
            )

    @classmethod
    def from_bytes(
        cls,
        bytes: bytes,
        cropping: Optional[Cropping] = None,
        controls: Optional[Sequence[ImageControl]] = None,
    ):
        image = base64.b64encode(bytes).decode()
        return cls(image, cropping, controls or [])

    @classmethod
    def from_url(cls, url: str, controls: Optional[Sequence[ImageControl]] = None):
        """
        Downloads a file and prepare it to be used in a prompt.
        The image will be [center cropped](https://pytorch.org/vision/stable/transforms.html#torchvision.transforms.CenterCrop)
        """
        return cls.from_bytes(
            cls._get_url(url), cropping=None, controls=controls or None
        )

    @classmethod
    def from_url_with_cropping(
        cls,
        url: str,
        upper_left_x: int,
        upper_left_y: int,
        crop_size: int,
        controls: Optional[Sequence[ImageControl]] = None,
    ):
        """
        Downloads a file and prepare it to be used in a prompt.
        upper_left_x, upper_left_y and crop_size are used to crop the image.
        """
        cropping = Cropping(
            upper_left_x=upper_left_x, upper_left_y=upper_left_y, size=crop_size
        )
        bytes = cls._get_url(url)
        return cls.from_bytes(bytes, cropping=cropping, controls=controls or [])

    @classmethod
    def from_file(
        cls, path: Union[str, Path], controls: Optional[Sequence[ImageControl]] = None
    ):
        """
        Load an image from disk and prepare it to be used in a prompt
        If they are not provided then the image will be [center cropped](https://pytorch.org/vision/stable/transforms.html#torchvision.transforms.CenterCrop)
        """
        with open(path, "rb") as f:
            image = f.read()
        return cls.from_bytes(image, None, controls or [])

    @classmethod
    def from_file_with_cropping(
        cls,
        path: str,
        upper_left_x: int,
        upper_left_y: int,
        crop_size: int,
        controls: Optional[Sequence[ImageControl]] = None,
    ):
        """
        Load an image from disk and prepare it to be used in a prompt
        upper_left_x, upper_left_y and crop_size are used to crop the image.
        """
        cropping = Cropping(
            upper_left_x=upper_left_x, upper_left_y=upper_left_y, size=crop_size
        )
        with open(path, "rb") as f:
            bytes = f.read()
        return cls.from_bytes(bytes, cropping=cropping, controls=controls or None)

    @classmethod
    def _get_url(cls, url: str) -> bytes:
        response = requests.get(url)
        response.raise_for_status()
        return response.content

    def to_json(self) -> Mapping[str, Any]:
        """
        A dict if serialized to JSON is suitable as a prompt element
        """
        if self.cropping is None:
            return {
                "type": "image",
                "data": self.base_64,
                "controls": [control.to_json() for control in self.controls],
            }
        else:
            return {
                "type": "image",
                "data": self.base_64,
                "x": self.cropping.upper_left_x,
                "y": self.cropping.upper_left_y,
                "size": self.cropping.size,
                "controls": [control.to_json() for control in self.controls],
            }

    @staticmethod
    def from_json(json: Mapping[str, Any]) -> "Image":
        return Image(base_64=json["data"], cropping=None, controls=[])

    def to_image(self) -> PILImage:
        return PIL.Image.open(io.BytesIO(base64.b64decode(self.base_64)))

    def dimensions(self) -> Tuple[int, int]:
        image = self.to_image()
        return (image.width, image.height)


PromptItem = Union[Text, Tokens, Image]


@dataclass
class Prompt:
    """
    Examples:
        >>> prompt = Prompt.from_text("Provide a short description of AI:")
        >>> prompt = Prompt([
                Image.from_url(url),
                Text.from_text("Provide a short description of AI:"),
            ])
    """

    items: Sequence[PromptItem]

    def __init__(self, items: Union[str, Sequence[PromptItem]]):
        if isinstance(items, str):
            items = [Text(items, [])]
        self.items = items

    @staticmethod
    def from_text(
        text: str, controls: Optional[Sequence[TextControl]] = None
    ) -> "Prompt":
        return Prompt([Text(text, controls or [])])

    @staticmethod
    def from_image(image: Image) -> "Prompt":
        return Prompt([image])

    @staticmethod
    def from_tokens(
        tokens: Sequence[int], controls: Optional[Sequence[TokenControl]] = None
    ) -> "Prompt":
        """
        Examples:
            >>> prompt = Prompt.from_tokens([1, 2, 3])
        """
        return Prompt([Tokens(tokens, controls or [])])

    def to_json(self) -> Sequence[Mapping[str, Any]]:
        return [_to_json(item) for item in self.items]

    @staticmethod
    def from_json(items_json: Sequence[Mapping[str, Any]]) -> "Prompt":
        return Prompt(
            [
                item
                for item in (_prompt_item_from_json(item) for item in items_json)
                if item
            ]
        )


def _prompt_item_from_json(item: Mapping[str, Any]) -> Optional[PromptItem]:
    item_type = item.get("type")
    if item_type == "text":
        return Text.from_json(item)
    if item_type == "image":
        return Image.from_json(item)
    if item_type == "token_ids":
        return Tokens.from_json(item)
    # Skip item instead of raising an error to prevent failures of old clients
    # when item types are extended
    return None


def _to_json(item: PromptItem) -> Mapping[str, Any]:
    if hasattr(item, "to_json"):
        return item.to_json()  # type: ignore
    # Required for backwards compatibility
    # item could be a plain piece of text or a plain list of token-ids
    elif isinstance(item, str):
        return {"type": "text", "data": item}
    elif isinstance(item, List):
        return {"type": "token_ids", "data": item}
    else:
        raise ValueError(
            "The item in the prompt is not valid. Try either a string or an Image."
        )



from re import finditer
from typing import Dict, Iterable, Mapping, NewType, Tuple, Union
from uuid import UUID, uuid4
from liquid import Template

from aleph_alpha_client.prompt import Image, Prompt, PromptItem, Text, Tokens

Placeholder = NewType("Placeholder", UUID)


class PromptTemplate:
    """Allows to build a `Prompt` using the `liquid template language <https://shopify.github.io/liquid/>`_.

    To add non-text prompt items first you have to save it to the template with the `template.placeholder()` function.
    To embed the items in the template, pass the placeholder in the place(s) where you would like the items.

    Example:
        >>> image = Image.from_file(Path("path-to-image"))
        >>> template = PromptTemplate(
            '''{%- for name in names -%}
            Hello {{name}}!
            {% endfor -%}
            {{ image }}
            ''')
        >>> placeholder = template.placeholder(image)
        >>> names = ["World", "Rutger"]
        >>> prompt = template.to_prompt(names=names, image=placeholder)
        >>> request = CompletionRequest(prompt=prompt)
    """

    def __init__(self, template_str: str) -> None:
        """Initialize with the liquid template string.

        Parameters:
            template_str: the liquid template string
        """
        self.template = Template(template_str)
        self.non_text_items: Dict[Placeholder, Union[Image, Tokens]] = {}

    def placeholder(self, prompt_item: Union[Image, Tokens]) -> Placeholder:
        """Saves a non-text prompt item to the template and returns a placeholder

        The placeholder is used to embed the prompt item in the template
        """
        id = Placeholder(uuid4())
        self.non_text_items[id] = prompt_item
        return id

    def _join_character(
        self, first_item: Union[Text, Image, Tokens, None], second_item: Text
    ) -> str:
        if (
            isinstance(first_item, Text)
            and not first_item.text[-1].isspace()
            and not second_item.text[0].isspace()
        ):
            return " "
        else:
            return ""

    def embed_prompt(self, prompt: Prompt) -> str:
        """Embeds a prompt in a prompt template

        Adds whitespace between text items if there is no whitespace between them.
        In case of non-text prompt items, this embeds them into the end result.

        Example:
            >>> user_prompt = Prompt(
                    [
                        Tokens.from_token_ids([1, 2, 3]),
                        Text.from_text("cool"),
                        Image.from_file(Path("path-to-image")),
                    ]
                )
            >>> template = PromptTemplate("Question: {{user_prompt}}\\n Answer: ")
            >>> prompt = template.to_prompt(user_prompt=template.embed_prompt(user_prompt))

        Parameters:
            prompt: prompt to embed in the template
        """
        prompt_text = ""
        last_item = None
        for item in prompt.items:
            if isinstance(item, Text):
                if len(item.text) == 0:
                    continue
                prompt_text = str.join(
                    self._join_character(last_item, item), [prompt_text, item.text]
                )
            else:
                prompt_text = str.join("", [prompt_text, str(self.placeholder(item))])
            last_item = item
        return prompt_text

    def to_prompt(self, **kwargs) -> Prompt:
        """Creates a `Prompt` from the template string and the given parameters.

        Provided parameters are passed to `liquid.Template.render`.
        """
        liquid_prompt: str = self.template.render(**kwargs)
        placeholder_indices = self._compute_indices(
            self.non_text_items.keys(), liquid_prompt
        )
        modalities = _modalities_from(
            placeholder_indices, self.non_text_items, liquid_prompt
        )

        self.non_text_items = {}
        return Prompt(list(modalities))

    def _compute_indices(
        self, placeholders: Iterable[Placeholder], template: str
    ) -> Iterable[Tuple[int, int]]:
        if not self.non_text_items:
            return []
        pattern = f"({'|'.join(str(placeholder) for placeholder in placeholders)})"
        return ((match.start(), match.end()) for match in finditer(pattern, template))


def _modalities_from(
    placeholder_indices: Iterable[Tuple[int, int]],
    prompt_items_by_placeholder: Mapping[Placeholder, Union[Image, Tokens]],
    template: str,
) -> Iterable[PromptItem]:
    last_to = 0
    for placeholder_from, placeholder_to in placeholder_indices:
        if last_to < placeholder_from:
            yield Text.from_text(template[last_to:placeholder_from])
        yield prompt_items_by_placeholder[
            Placeholder(UUID(template[placeholder_from:placeholder_to]))
        ]
        last_to = placeholder_to
    if last_to < len(template):
        yield Text.from_text(template[last_to:])



from dataclasses import asdict, dataclass
from typing import Any, Mapping, Optional, Union, Type
from pydantic import BaseModel


@dataclass(frozen=True)
class JSONSchema:
    """
    JSON schema that structured output must adhere to.

    Parameters:
        schema:
            JSON schema that structured output must adhere to.
        name:
            Name of the schema.
        description:
            Description of the schema.
        strict:
            Whether the schema should be strictly enforced.

    Examples:
        >>> schema = JSONSchema(
        >>>     schema={
        >>>         'type': 'object',
        >>>         'title': 'Aquarium',
        >>>         'properties': {
        >>>             'nemo': {
        >>>                 'type': 'string',
        >>>                 'title': 'Nemo'
        >>>             }
        >>>         },
        >>>         'required': ['nemo']
        >>>     },
        >>>     name="aquarium",
        >>>     description="Describe nemo",
        >>>     strict=True
        >>> )
    """

    schema: Mapping[str, Any]
    name: str
    description: Optional[str] = None
    strict: Optional[bool] = False

    def to_json(self) -> Mapping[str, Any]:
        return {"type": "json_schema", "json_schema": asdict(self)}

    @classmethod
    def from_pydantic(cls, model_class: Type[BaseModel]) -> "JSONSchema":
        """
        Create a JSONSchema from a Pydantic model class.

        Parameters:
            model_class: A Pydantic BaseModel class

        Returns:
            JSONSchema instance with the schema generated from the Pydantic model

        Raises:
            ValueError: If the provided class is not a Pydantic BaseModel
        """

        schema = model_class.model_json_schema()
        name = getattr(model_class, "__name__", "generated_schema")

        # Use the name as default if no description is provided
        description = getattr(model_class, "__doc__", name)

        return cls(schema=schema, name=name, description=description, strict=True)


ResponseFormat = Union[JSONSchema, Type[BaseModel]]



from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    List,
    Dict,
    Mapping,
    Optional,
    Union,
)

from aleph_alpha_client import Text

from aleph_alpha_client.prompt import ControlTokenOverlap, Image, Prompt, PromptItem


class ExplanationPostprocessing(Enum):
    """
    Available types of explanation postprocessing.

    Square:
        Square each score
    Absolute:
        Take the absolute value of each score
    """

    Square = "square"
    Absolute = "absolute"

    def to_json(self) -> str:
        return self.value


@dataclass(frozen=True)
class CustomGranularity:
    """
    Allows for passing a custom delimiter to determine the granularity to
    to explain the prompt by. The text of the prompt will be split by the
    delimiter you provide.

    Parameters:
        delimiter (str, required):
            String to split the text in the prompt by for generating
            explanations for your prompt.
    """

    delimiter: str

    def to_json(self) -> Mapping[str, Any]:
        return {"type": "custom", "delimiter": self.delimiter}


class PromptGranularity(Enum):
    Token = "token"
    Word = "word"
    Sentence = "sentence"
    Paragraph = "paragraph"

    def to_json(self) -> Mapping[str, Any]:
        return {"type": self.value}


def prompt_granularity_to_json(
    prompt_granularity: Union[PromptGranularity, str, CustomGranularity],
) -> Mapping[str, Any]:
    # we allow str for backwards compatibility
    # This was previously possible because PromptGranularity was not an Enum
    if isinstance(prompt_granularity, str):
        return {"type": prompt_granularity}

    return prompt_granularity.to_json()


class TargetGranularity(Enum):
    """
    How many explanations should be returned in the output.

    Complete:
        Return one explanation for the entire target. Helpful in many cases to determine which parts of the prompt contribute overall to the given completion.
    Token:
        Return one explanation for each token in the target.
    """

    Complete = "complete"
    Token = "token"

    def to_json(self) -> str:
        return self.value


@dataclass(frozen=True)
class ExplanationRequest:
    """
    Describes an Explanation request you want to make agains the API.

    Parameters:
        prompt (Prompt, required)
            Prompt you want to generate explanations for a target completion.
        target (str, required)
            The completion string to be explained based on model probabilities.
        contextual_control_threshold (float, default None)
            If set to None, attention control parameters only apply to those tokens that have
            explicitly been set in the request.
            If set to a non-None value, we apply the control parameters to similar tokens as well.
            Controls that have been applied to one token will then be applied to all other tokens
            that have at least the similarity score defined by this parameter.
            The similarity score is the cosine similarity of token embeddings.
        control_factor (float, default None):
            The amount to adjust model attention by.
            For Explanation, you want to supress attention, and the API will default to 0.1.
            Values between 0 and 1 will supress attention.
            A value of 1 will have no effect.
            Values above 1 will increase attention.
        control_token_overlap (ControlTokenOverlap, default None)
            What to do if a control partially overlaps with a text or image token.
            If set to "partial", the factor will be adjusted proportionally with the amount
            of the token it overlaps. So a factor of 2.0 of a control that only covers 2 of
            4 token characters, would be adjusted to 1.5.
            If set to "complete", the full factor will be applied as long as the control
            overlaps with the token at all.
        control_log_additive (bool, default None)
            True: apply control by adding the log(control_factor) to attention scores.
            False: apply control by (attention_scores - - attention_scores.min(-1)) * control_factor
            If None, the API will default to True
        prompt_granularity (Union[PromptGranularity, str, CustomGranularity], default None)
            At which granularity should the target be explained in terms of the prompt.
            If you choose, for example, "sentence" then we report the importance score of each
            sentence in the prompt towards generating the target output.

            If you do not choose a granularity then we will try to find the granularity that
            brings you closest to around 30 explanations. For large documents, this would likely
            be sentences. For short prompts this might be individual words or even tokens.

            If you choose a custom granularity then you must provide a custom delimiter. We then
            split your prompt by that delimiter. This might be helpful if you are using few-shot
            prompts that contain stop sequences.

            We currently support providing the prompt_granularity as PromptGranularity (recommended)
            or CustomGranularity (if needed) or str (deprecated). Note that supplying plain strings
            only makes sense if you choose one of the values defined in the PromptGranularity enum.
            All other strings will be rejected by the API. In future versions we might cut support
            for plain str values.

            For image prompt items, the granularities determine into how many tiles we divide
            the image for the explanation.
            "token" -> 12x12
            "word" -> 6x6
            "sentence" -> 3x3
            "paragraph" -> 1
        target_granularity (TargetGranularity, default None)
            How many explanations should be returned in the output.

            "complete" -> Return one explanation for the entire target. Helpful in many cases to determine which parts of the prompt contribute overall to the given completion.
            "token" -> Return one explanation for each token in the target.

            If None, API will default to "complete"
        postprocessing (ExplanationPostprocessing, default None)
            Optionally apply postprocessing to the difference in cross entropy scores for each token.
            "none": Apply no postprocessing.
            "absolute": Return the absolute value of each value.
            "square": Square each value
        normalize (bool, default None)
            Return normalized scores. Minimum score becomes 0 and maximum score becomes 1. Applied after any postprocessing
    """

    prompt: Prompt
    target: str
    contextual_control_threshold: Optional[float] = None
    control_factor: Optional[float] = None
    control_token_overlap: Optional[ControlTokenOverlap] = None
    control_log_additive: Optional[bool] = None
    prompt_granularity: Optional[Union[PromptGranularity, str, CustomGranularity]] = (
        None
    )
    target_granularity: Optional[TargetGranularity] = None
    postprocessing: Optional[ExplanationPostprocessing] = None
    normalize: Optional[bool] = None

    def to_json(self) -> Mapping[str, Any]:
        payload: Dict[str, Any] = {
            "prompt": self.prompt.to_json(),
            "target": self.target,
        }
        if self.contextual_control_threshold is not None:
            payload["contextual_control_threshold"] = self.contextual_control_threshold
        if self.control_token_overlap is not None:
            payload["control_token_overlap"] = self.control_token_overlap.to_json()
        if self.postprocessing is not None:
            payload["postprocessing"] = self.postprocessing.to_json()
        if self.control_log_additive is not None:
            payload["control_log_additive"] = self.control_log_additive
        if self.prompt_granularity is not None:
            payload["prompt_granularity"] = prompt_granularity_to_json(
                self.prompt_granularity
            )
        if self.target_granularity is not None:
            payload["target_granularity"] = self.target_granularity.to_json()
        if self.postprocessing is not None:
            payload["postprocessing"] = self.postprocessing.to_json()
        if self.normalize is not None:
            payload["normalize"] = self.normalize
        if self.control_factor is not None:
            payload["control_factor"] = self.control_factor

        return payload


@dataclass(frozen=True)
class TextScore:
    start: int
    length: int
    score: float

    @staticmethod
    def from_json(score: Any) -> "TextScore":
        return TextScore(
            start=score["start"],
            length=score["length"],
            score=score["score"],
        )


@dataclass(frozen=True)
class TextScoreWithRaw:
    start: int
    length: int
    score: float
    text: str

    @staticmethod
    def from_text_score(score: TextScore, prompt: Text) -> "TextScoreWithRaw":
        return TextScoreWithRaw(
            start=score.start,
            length=score.length,
            score=score.score,
            text=prompt.text[score.start : score.start + score.length],
        )


@dataclass(frozen=True)
class ImageScore:
    left: float
    top: float
    width: float
    height: float
    score: float

    @staticmethod
    def from_json(score: Any) -> "ImageScore":
        return ImageScore(
            left=score["rect"]["left"],
            top=score["rect"]["top"],
            width=score["rect"]["width"],
            height=score["rect"]["height"],
            score=score["score"],
        )


@dataclass(frozen=True)
class TargetScore:
    start: int
    length: int
    score: float

    @staticmethod
    def from_json(score: Any) -> "TargetScore":
        return TargetScore(
            start=score["start"],
            length=score["length"],
            score=score["score"],
        )


@dataclass(frozen=True)
class TargetScoreWithRaw:
    start: int
    length: int
    score: float
    text: str

    @staticmethod
    def from_target_score(score: TargetScore, target: str) -> "TargetScoreWithRaw":
        return TargetScoreWithRaw(
            start=score.start,
            length=score.length,
            score=score.score,
            text=target[score.start : score.start + score.length],
        )


@dataclass(frozen=True)
class TokenScore:
    score: float

    @staticmethod
    def from_json(score: Any) -> "TokenScore":
        return TokenScore(
            score=score,
        )


@dataclass(frozen=True)
class ImagePromptItemExplanation:
    """
    Explains the importance of an image prompt item.
    The amount of items in the "scores" array depends on the granularity setting.
    Each score object contains the top-left corner of a rectangular area in the image prompt.
    The coordinates are all between 0 and 1 in terms of the total image size
    """

    scores: List[ImageScore]

    @staticmethod
    def from_json(item: Dict[str, Any]) -> "ImagePromptItemExplanation":
        return ImagePromptItemExplanation(
            scores=[ImageScore.from_json(score) for score in item["scores"]]
        )

    def in_pixels(self, prompt_item: PromptItem) -> "ImagePromptItemExplanation":
        if not isinstance(prompt_item, Image):
            raise ValueError
        (original_image_width, original_image_height) = prompt_item.dimensions()
        return ImagePromptItemExplanation(
            [
                ImageScore(
                    left=int(score.left * original_image_width),
                    width=int(score.width * original_image_width),
                    top=int(score.top * original_image_height),
                    height=int(score.height * original_image_height),
                    score=score.score,
                )
                for score in self.scores
            ]
        )


@dataclass(frozen=True)
class TextPromptItemExplanation:
    """
    Explains the importance of a text prompt item.
    The amount of items in the "scores" array depends on the granularity setting.
    Each score object contains an inclusive start character and a length of the substring plus
    a floating point score value.
    """

    scores: List[Union[TextScore, TextScoreWithRaw]]

    @staticmethod
    def from_json(item: Dict[str, Any]) -> "TextPromptItemExplanation":
        return TextPromptItemExplanation(
            scores=[TextScore.from_json(score) for score in item["scores"]]
        )

    def with_text(self, prompt: Text) -> "TextPromptItemExplanation":
        return TextPromptItemExplanation(
            scores=[
                (
                    TextScoreWithRaw.from_text_score(score, prompt)
                    if isinstance(score, TextScore)
                    else score
                )
                for score in self.scores
            ]
        )


@dataclass(frozen=True)
class TargetPromptItemExplanation:
    """
    Explains the importance of text in the target string that came before the currently
    to-be-explained target token. The amount of items in the "scores" array depends on the
    granularity setting.
    Each score object contains an inclusive start character and a length of the substring plus
    a floating point score value.
    """

    scores: List[Union[TargetScore, TargetScoreWithRaw]]

    @staticmethod
    def from_json(item: Dict[str, Any]) -> "TargetPromptItemExplanation":
        return TargetPromptItemExplanation(
            scores=[TargetScore.from_json(score) for score in item["scores"]]
        )

    def with_text(self, prompt: str) -> "TargetPromptItemExplanation":
        return TargetPromptItemExplanation(
            scores=[
                (
                    TargetScoreWithRaw.from_target_score(score, prompt)
                    if isinstance(score, TargetScore)
                    else score
                )
                for score in self.scores
            ]
        )


@dataclass(frozen=True)
class TokenPromptItemExplanation:
    """Explains the importance of a request prompt item of type "token_ids".
    Will contain one floating point importance value for each token in the same order as in the original prompt.
    """

    scores: List[TokenScore]

    @staticmethod
    def from_json(item: Dict[str, Any]) -> "TokenPromptItemExplanation":
        return TokenPromptItemExplanation(
            scores=[TokenScore.from_json(score) for score in item["scores"]]
        )


@dataclass(frozen=True)
class Explanation:
    """
    Explanations for a given portion of the target.

    Parameters:
        target (str, required)
            If target_granularity was set to "complete", then this will be the entire target. If it was set to "token", this will be a single target token.
        items (List[Union[TextPromptItemExplanation, TargetPromptItemExplanation, TokenPromptItemExplanation, ImagePromptItemExplanation], required)
            Contains one item for each prompt item (in order), and the last item refers to the target.
    """

    target: str
    items: List[
        Union[
            TextPromptItemExplanation,
            TargetPromptItemExplanation,
            TokenPromptItemExplanation,
            ImagePromptItemExplanation,
        ]
    ]

    def prompt_item_from_json(
        item: Any,
    ) -> Union[
        TextPromptItemExplanation,
        ImagePromptItemExplanation,
        TargetPromptItemExplanation,
        TokenPromptItemExplanation,
    ]:
        if item["type"] == "text":
            return TextPromptItemExplanation.from_json(item)
        elif item["type"] == "target":
            return TargetPromptItemExplanation.from_json(item)
        elif item["type"] == "image":
            return ImagePromptItemExplanation.from_json(item)
        elif item["type"] == "token_ids":
            return TokenPromptItemExplanation.from_json(item)
        else:
            raise NotImplementedError("Unsupported explanation type")

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "Explanation":
        return Explanation(
            target=json["target"],
            items=[Explanation.prompt_item_from_json(item) for item in json["items"]],
        )

    def with_image_prompt_items_in_pixels(self, prompt: Prompt) -> "Explanation":
        return Explanation(
            target=self.target,
            items=[
                (
                    item.in_pixels(prompt.items[item_index])
                    if isinstance(item, ImagePromptItemExplanation)
                    else item
                )
                for item_index, item in enumerate(self.items)
            ],
        )

    def with_text_from_prompt(self, prompt: Prompt, target: str) -> "Explanation":
        items: List[
            Union[
                TextPromptItemExplanation,
                ImagePromptItemExplanation,
                TargetPromptItemExplanation,
                TokenPromptItemExplanation,
            ]
        ] = []
        for item_index, item in enumerate(self.items):
            if isinstance(item, TextPromptItemExplanation):
                # separate variable to fix linting error
                prompt_item = prompt.items[item_index]
                if isinstance(prompt_item, Text):
                    items.append(item.with_text(prompt_item))
                else:
                    items.append(item)
            elif isinstance(item, TargetPromptItemExplanation):
                items.append(item.with_text(target))
            else:
                items.append(item)
        return Explanation(
            target=self.target,
            items=items,
        )


@dataclass(frozen=True)
class ExplanationResponse:
    """
    The top-level response data structure that will be returned from an explanation request.

    Parameters:
        model_version (str, required)
            Version of the model used to generate the explanation.
        explanations (List[Explanation], required)
            This array will contain one explanation object for each portion of the target.
    """

    model_version: str
    explanations: List[Explanation]

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "ExplanationResponse":
        return ExplanationResponse(
            model_version=json["model_version"],
            explanations=[
                Explanation.from_json(explanation)
                for explanation in json["explanations"]
            ],
        )

    def with_image_prompt_items_in_pixels(
        self, prompt: Prompt
    ) -> "ExplanationResponse":
        mapped_explanations = [
            explanation.with_image_prompt_items_in_pixels(prompt)
            for explanation in self.explanations
        ]
        return ExplanationResponse(self.model_version, mapped_explanations)

    def with_text_from_prompt(
        self, request: ExplanationRequest
    ) -> "ExplanationResponse":
        mapped_explanations = [
            explanation.with_text_from_prompt(request.prompt, request.target)
            for explanation in self.explanations
        ]
        return ExplanationResponse(self.model_version, mapped_explanations)



from aleph_alpha_client.aleph_alpha_client import AsyncClient, Client
from aleph_alpha_client.tokenization import TokenizationRequest
from aleph_alpha_client.detokenization import DetokenizationRequest
import pytest

# AsyncClient


@pytest.mark.vcr
async def test_can_tokenize_with_async_client(
    async_client: AsyncClient, model_name: str
):
    request = TokenizationRequest(prompt="hello", token_ids=True, tokens=True)

    response = await async_client.tokenize(request, model=model_name)
    assert response.tokens and len(response.tokens) == 1
    assert response.token_ids and len(response.token_ids) == 1


# Client


@pytest.mark.vcr
def test_tokenize(sync_client: Client, model_name: str):
    response = sync_client.tokenize(
        request=TokenizationRequest("Hello", tokens=True, token_ids=True),
        model=model_name,
    )

    assert response.tokens and len(response.tokens) == 1
    assert response.token_ids and len(response.token_ids) == 1


@pytest.mark.vcr
def test_offline_tokenize_sync(sync_client: Client, model_name: str):
    prompt = "Hello world"

    tokenizer = sync_client.tokenizer(model_name)
    offline_tokenization = tokenizer.encode(prompt)

    tokenization_request = TokenizationRequest(
        prompt=prompt, token_ids=True, tokens=True
    )
    online_tokenization_response = sync_client.tokenize(
        tokenization_request, model_name
    )

    assert offline_tokenization.ids == online_tokenization_response.token_ids
    assert offline_tokenization.tokens == online_tokenization_response.tokens


@pytest.mark.vcr
def test_offline_detokenize_sync(sync_client: Client, model_name: str):
    prompt = "Hello world"

    tokenizer = sync_client.tokenizer(model_name)
    offline_tokenization = tokenizer.encode(prompt)
    offline_detokenization = tokenizer.decode(offline_tokenization.ids)

    detokenization_request = DetokenizationRequest(token_ids=offline_tokenization.ids)
    online_detokenization_response = sync_client.detokenize(
        detokenization_request, model_name
    )

    assert offline_detokenization == online_detokenization_response.result


@pytest.mark.vcr
async def test_offline_tokenizer_async(async_client: AsyncClient, model_name: str):
    prompt = "Hello world"

    tokenizer = await async_client.tokenizer(model_name)
    offline_tokenization = tokenizer.encode(prompt)

    request = TokenizationRequest(prompt=prompt, token_ids=True, tokens=True)
    response = await async_client.tokenize(request, model_name)

    assert offline_tokenization.ids == response.token_ids
    assert offline_tokenization.tokens == response.tokens



from dataclasses import dataclass
from typing import Any, Dict, List, Mapping, Optional, Sequence


@dataclass(frozen=True)
class RerankRequest:
    """Request for reranking documents against a query.

    This endpoint takes in a query and a list of documents and produces an array
    with each document assigned a relevance score.

    Parameters:
        query (str, required):
            The query to rerank the documents against.

        documents (Sequence[str], required):
            The list of documents to rerank.

        top_n (int, optional):
            The number of documents to return. Defaults to the number of documents
            if not provided.

    Examples:
        >>> request = RerankRequest(
        ...     query="What is the capital of France?",
        ...     documents=[
        ...         "The capital of Brazil is Brasilia.",
        ...         "The capital of France is Paris.",
        ...         "Horses and cows are both animals.",
        ...     ],
        ...     top_n=2,
        ... )
        >>> response = client.rerank(request, model="your-reranker-model")
    """

    query: str
    documents: Sequence[str]
    top_n: Optional[int] = None

    def to_json(self) -> Mapping[str, Any]:
        """Convert the request to a JSON-serializable dictionary."""
        json_request: Dict[str, Any] = {
            "query": self.query,
            "documents": list(self.documents),
        }
        if self.top_n is not None:
            json_request["top_n"] = self.top_n
        return json_request


@dataclass(frozen=True)
class RerankResult:
    """A single reranked document result.

    Parameters:
        index (int):
            The index of the document in the original list of documents.

        relevance_score (float):
            The relevance score of the document.
    """

    index: int
    relevance_score: float

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "RerankResult":
        return RerankResult(
            index=json["index"],
            relevance_score=json["relevance_score"],
        )


@dataclass(frozen=True)
class RerankUsage:
    """Usage statistics for the rerank request.

    Parameters:
        completion_tokens (int):
            Number of tokens in the generated completion. Will always be 0 for rerank tasks.

        prompt_tokens (int):
            Number of tokens in the prompt. Will always be 0 for rerank tasks.

        total_tokens (int):
            Total number of tokens used in the request.
    """

    completion_tokens: int
    prompt_tokens: int
    total_tokens: int

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "RerankUsage":
        return RerankUsage(
            completion_tokens=json["completion_tokens"],
            prompt_tokens=json["prompt_tokens"],
            total_tokens=json["total_tokens"],
        )


@dataclass(frozen=True)
class RerankResponse:
    """Response from a rerank request.

    Parameters:
        results (List[RerankResult]):
            The reranked results, each containing the original document index
            and its relevance score.

        usage (RerankUsage):
            Usage statistics for the request.
    """

    results: List[RerankResult]
    usage: RerankUsage

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "RerankResponse":
        return RerankResponse(
            results=[RerankResult.from_json(r) for r in json["results"]],
            usage=RerankUsage.from_json(json["usage"]),
        )



from dataclasses import asdict, dataclass
from typing import Any, Dict, Mapping


@dataclass(frozen=True)
class SteeringPairedExample:
    negative: str
    positive: str

    def to_json(self) -> Mapping[str, Any]:
        return self._asdict()

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class SteeringConceptCreationRequest:
    """
    Creates a new steering concept that can be used in completion requests.

    Parameters:
        examples
            A list of SteeringPairedExample objects.

    Examples:
        >>> examples = [
        >>>     SteeringPairedExample(
        >>>         negative="I appreciate your valuable feedback on this matter.",
        >>>         positive="Thanks for the real talk, fam.",
        >>>     ),
        >>>     SteeringPairedExample(
        >>>         negative="The financial projections indicate significant growth potential.",
        >>>         positive="Yo, these numbers are looking mad stacked!",
        >>>     ),
        >>> ]
        >>> SteeringConceptCreationRequest(examples=examples)
    """

    examples: list[SteeringPairedExample]

    def to_json(self) -> Mapping[str, Any]:
        return {
            **self._asdict(),
            "examples": [e.to_json() for e in self.examples],
        }

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class SteeringConceptCreationResponse:
    id: str

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "SteeringConceptCreationResponse":
        return SteeringConceptCreationResponse(
            id=json["id"],
        )



from dataclasses import asdict, dataclass
from enum import Enum
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Union,
    Optional,
    Sequence,
    Tuple,
    Literal,
)
from aleph_alpha_client.prompt import Prompt


@dataclass(frozen=True)
class EmbeddingRequest:
    """
    Embeds a text and returns vectors that can be used for downstream tasks (e.g. semantic similarity) and models (e.g. classifiers).

    Parameters:
        prompt
            The text and/or image(s) to be embedded.
        layers
            A list of layer indices from which to return embeddings.

            * Index 0 corresponds to the word embeddings used as input to the first transformer layer
            * Index 1 corresponds to the hidden state as output by the first transformer layer, index 2 to the output of the second layer etc.
            * Index -1 corresponds to the last transformer layer (not the language modelling head), index -2 to the second last layer etc.
        pooling
            Pooling operation to use.
            Pooling operations include:

            * mean: aggregate token embeddings across the sequence dimension using an average
            * max: aggregate token embeddings across the sequence dimension using a maximum
            * last_token: just use the last token
            * abs_max: aggregate token embeddings across the sequence dimension using a maximum of absolute values
        type
            Type of the embedding (e.g. symmetric or asymmetric)
        tokens
            Flag indicating whether the tokenized prompt is to be returned (True) or not (False)
        normalize
            Return normalized embeddings. This can be used to save on additional compute when applying a cosine similarity metric.
        contextual_control_threshold (float, default None)
            If set to None, attention control parameters only apply to those tokens that have
            explicitly been set in the request.
            If set to a non-None value, we apply the control parameters to similar tokens as well.
            Controls that have been applied to one token will then be applied to all other tokens
            that have at least the similarity score defined by this parameter.
            The similarity score is the cosine similarity of token embeddings.
        control_log_additive (bool, default True)
            True: apply control by adding the log(control_factor) to attention scores.
            False: apply control by (attention_scores - - attention_scores.min(-1)) * control_factor

    Examples:
        >>> prompt = Prompt.from_text("This is an example.")
        >>> EmbeddingRequest(prompt=prompt, layers=[-1], pooling=["mean"])
    """

    prompt: Prompt
    layers: List[int]
    pooling: List[str]
    type: Optional[str] = None
    tokens: bool = False
    normalize: bool = False
    contextual_control_threshold: Optional[float] = None
    control_log_additive: Optional[bool] = True

    def to_json(self) -> Mapping[str, Any]:
        return {
            **self._asdict(),
            "prompt": self.prompt.to_json(),
        }

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class EmbeddingResponse:
    model_version: str
    num_tokens_prompt_total: int
    embeddings: Optional[Dict[Tuple[str, str], List[float]]]
    tokens: Optional[List[str]]
    message: Optional[str] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "EmbeddingResponse":
        return EmbeddingResponse(
            model_version=json["model_version"],
            embeddings={
                (layer, pooling): embedding
                for layer, pooling_dict in json["embeddings"].items()
                for pooling, embedding in pooling_dict.items()
            },
            tokens=json.get("tokens"),
            message=json.get("message"),
            num_tokens_prompt_total=json["num_tokens_prompt_total"],
        )


class SemanticRepresentation(Enum):
    """
    Available types of semantic representations that prompts can be embedded with.

    Symmetric:
        `Symmetric` is useful for comparing prompts to each other, in use cases such as clustering, classification, similarity, etc. `Symmetric` embeddings should be compared with other `Symmetric` embeddings.
    Document:
        `Document` and `Query` are used together in use cases such as search where you want to compare shorter queries against larger documents.

        `Document` embeddings are optimized for larger pieces of text to compare queries against.
    Query:
        `Document` and `Query` are used together in use cases such as search where you want to compare shorter queries against larger documents.

        `Query` embeddings are optimized for shorter texts, such as questions or keywords.
    """

    Symmetric = "symmetric"
    Document = "document"
    Query = "query"


@dataclass(frozen=True)
class SemanticEmbeddingRequest:
    """
    Embeds a text and returns vectors that can be used for downstream tasks (e.g. semantic similarity) and models (e.g. classifiers).

    Parameters:
        prompt
            The text and/or image(s) to be embedded.
        representation
            Semantic representation to embed the prompt with.
        compress_to_size
            Options available: 128

            The default behavior is to return the full embedding, but you can optionally request an embedding compressed to a smaller set of dimensions.

            Full embedding sizes for supported models:
              - luminous-base: 5120

            The 128 size is expected to have a small drop in accuracy performance (4-6%), with the benefit of being much smaller, which makes comparing these embeddings much faster for use cases where speed is critical.

            The 128 size can also perform better if you are embedding really short texts or documents.
        normalize
            Return normalized embeddings. This can be used to save on additional compute when applying a cosine similarity metric.
        contextual_control_threshold (float, default None)
            If set to None, attention control parameters only apply to those tokens that have
            explicitly been set in the request.
            If set to a non-None value, we apply the control parameters to similar tokens as well.
            Controls that have been applied to one token will then be applied to all other tokens
            that have at least the similarity score defined by this parameter.
            The similarity score is the cosine similarity of token embeddings.
        control_log_additive (bool, default True)
            True: apply control by adding the log(control_factor) to attention scores.
            False: apply control by (attention_scores - - attention_scores.min(-1)) * control_factor

    Examples
        >>> texts = [
                "deep learning",
                "artificial intelligence",
                "deep diving",
                "artificial snow",
            ]
        >>> # Texts to compare
        >>> embeddings = []
        >>> for text in texts:
                request = SemanticEmbeddingRequest(prompt=Prompt.from_text(text), representation=SemanticRepresentation.Symmetric)
                result = model.semantic_embed(request)
                embeddings.append(result.embedding)
    """

    prompt: Prompt
    representation: SemanticRepresentation
    compress_to_size: Optional[int] = None
    normalize: bool = False
    contextual_control_threshold: Optional[float] = None
    control_log_additive: Optional[bool] = True

    def to_json(self) -> Mapping[str, Any]:
        return {
            **self._asdict(),
            "representation": self.representation.value,
            "prompt": self.prompt.to_json(),
        }

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class BatchSemanticEmbeddingRequest:
    """
    Embeds multiple multi-modal prompts and returns their embeddings in the same order as they were supplied.

    Parameters:
        prompts
            A list of texts and/or images to be embedded.
        representation
            Semantic representation to embed the prompt with.
        compress_to_size
            Options available: 128

            The default behavior is to return the full embedding, but you can optionally request an embedding compressed to a smaller set of dimensions.

            Full embedding sizes for supported models:
              - luminous-base: 5120

            The 128 size is expected to have a small drop in accuracy performance (4-6%), with the benefit of being much smaller, which makes comparing these embeddings much faster for use cases where speed is critical.

            The 128 size can also perform better if you are embedding really short texts or documents.
        normalize
            Return normalized embeddings. This can be used to save on additional compute when applying a cosine similarity metric.
        contextual_control_threshold (float, default None)
            If set to None, attention control parameters only apply to those tokens that have
            explicitly been set in the request.
            If set to a non-None value, we apply the control parameters to similar tokens as well.
            Controls that have been applied to one token will then be applied to all other tokens
            that have at least the similarity score defined by this parameter.
            The similarity score is the cosine similarity of token embeddings.
        control_log_additive (bool, default True)
            True: apply control by adding the log(control_factor) to attention scores.
            False: apply control by (attention_scores - - attention_scores.min(-1)) * control_factor

    Examples
        >>> texts = [
                "deep learning",
                "artificial intelligence",
                "deep diving",
                "artificial snow",
            ]
        >>> # Texts to compare
        >>> request = BatchSemanticEmbeddingRequest(prompts=[Prompt.from_text(text) for text in texts], representation=SemanticRepresentation.Symmetric)
            result = model.batch_semantic_embed(request)
    """

    prompts: Sequence[Prompt]
    representation: SemanticRepresentation
    compress_to_size: Optional[int] = None
    normalize: bool = False
    contextual_control_threshold: Optional[float] = None
    control_log_additive: Optional[bool] = True

    def to_json(self) -> Mapping[str, Any]:
        return {
            **self._asdict(),
            "representation": self.representation.value,
            "prompts": [prompt.to_json() for prompt in self.prompts],
        }

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class InstructableEmbeddingRequest:
    """
    Embeds a text and returns vectors that can be used for classification according to a given instruction.

    Parameters:
        input
            The text and/or image(s) to be embedded.
        instruction
            An instruction specifying the aspect to attend to when generating the embedding.
        normalize
            Return normalized embeddings. This can be used to save on additional compute when applying a cosine similarity metric.
        contextual_control_threshold (float, default None)
            If set to None, attention control parameters only apply to those tokens that have
            explicitly been set in the request.
            If set to a non-None value, we apply the control parameters to similar tokens as well.
            Controls that have been applied to one token will then be applied to all other tokens
            that have at least the similarity score defined by this parameter.
            The similarity score is the cosine similarity of token embeddings.
        control_log_additive (bool, default True)
            True: apply control by adding the log(control_factor) to attention scores.
            False: apply control by (attention_scores - - attention_scores.min(-1)) * control_factor

    Examples
        >>> texts = [
                "deep learning",
                "artificial intelligence",
                "deep diving",
                "artificial snow",
            ]
        >>> # Texts to compare
        >>> embeddings = []
        >>> for text in texts:
                request = InstructableEmbeddingRequest(
                    input=Prompt.from_text(text),
                    instruction="Represent the text to query a database of technical concepts",
                )
                result = model.instructable_embed(request)
                embeddings.append(result.embedding)
    """

    input: Prompt
    instruction: str
    normalize: bool = False
    contextual_control_threshold: Optional[float] = None
    control_log_additive: Optional[bool] = True

    def to_json(self) -> Mapping[str, Any]:
        return {
            **self._asdict(),
            "input": self.input.to_json(),
        }

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


EmbeddingVector = List[float]


@dataclass(frozen=True)
class SemanticEmbeddingResponse:
    """
    Response of a semantic embedding request

    Parameters:
        model_version
            Model name and version (if any) of the used model for inference
        embedding
            A list of floats that can be used to compare against other embeddings.
        message
            This field is no longer used.
    """

    model_version: str
    embedding: EmbeddingVector
    num_tokens_prompt_total: int
    message: Optional[str] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "SemanticEmbeddingResponse":
        return SemanticEmbeddingResponse(
            model_version=json["model_version"],
            embedding=json["embedding"],
            message=json.get("message"),
            num_tokens_prompt_total=json["num_tokens_prompt_total"],
        )


@dataclass(frozen=True)
class BatchSemanticEmbeddingResponse:
    """
    Response of a batch semantic embedding request

    Parameters:
        model_version
            Model name and version (if any) of the used model for inference
        embeddings
            A list of embeddings.
    """

    model_version: str
    embeddings: Sequence[EmbeddingVector]
    num_tokens_prompt_total: int

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "BatchSemanticEmbeddingResponse":
        return BatchSemanticEmbeddingResponse(
            model_version=json["model_version"],
            embeddings=json["embeddings"],
            num_tokens_prompt_total=json["num_tokens_prompt_total"],
        )

    def to_json(self) -> Mapping[str, Any]:
        return {
            **asdict(self),
            "embeddings": [embedding for embedding in self.embeddings],
        }

    @staticmethod
    def _from_model_version_and_embeddings(
        model_version: str,
        embeddings: Sequence[EmbeddingVector],
        num_tokens_prompt_total: int,
    ) -> "BatchSemanticEmbeddingResponse":
        return BatchSemanticEmbeddingResponse(
            model_version=model_version,
            embeddings=embeddings,
            num_tokens_prompt_total=num_tokens_prompt_total,
        )


@dataclass(frozen=True)
class InstructableEmbeddingResponse:
    """
    Response of an instructable embedding request

    Parameters:
        model_version
            Model name and version (if any) of the used model for inference
        embedding
            A list of floats that can be used to compare against other embeddings.
    """

    model_version: str
    embedding: EmbeddingVector
    num_tokens_prompt_total: int
    message: Optional[str] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "InstructableEmbeddingResponse":
        return InstructableEmbeddingResponse(
            model_version=json["model_version"],
            embedding=json["embedding"],
            num_tokens_prompt_total=json["num_tokens_prompt_total"],
        )


@dataclass(frozen=True)
class EmbeddingV2Request:
    """
    Embeds a text and returns embedding vectors following the OpenAI API specs.

    Parameters:
        input
            The input to be embedded. Can be a string, list of strings, list of integers (i.e.
            tokens) or a list of lists of integers (i.e. multiple token strings).
        dimensions
            The number of dimensions the resulting output embeddings should have. Note, not all
            models support this parameter.

    Examples
        >>> request = EmbeddingV2Request(
                input="Hello World!",
                dimensions=64,
            )
            result = model.embeddings(request)
    """

    input: Union[str, List[str], List[int], List[List[int]]]
    dimensions: Optional[int] = None
    encoding_format: Optional[Literal["float", "base64"]] = None

    def to_json(self) -> Mapping[str, Any]:
        body: Dict[str, Any] = {"input": self.input}
        if self.dimensions is not None:
            body["dimensions"] = self.dimensions
        if self.encoding_format is not None:
            body["encoding_format"] = self.encoding_format
        return body


@dataclass(frozen=True)
class Usage:
    """
    Usage statistics for the embedding request.
    """

    prompt_tokens: int
    total_tokens: int

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "Usage":
        return Usage(
            prompt_tokens=json["prompt_tokens"],
            total_tokens=json["total_tokens"],
        )


@dataclass(frozen=True)
class EmbeddingV2ResponseData:
    """
    Data structure for the embedding response in OpenAI compatible format.
    """

    object: str
    embedding: List[float]
    index: int

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "EmbeddingV2ResponseData":
        return EmbeddingV2ResponseData(
            object=json["object"],
            embedding=json["embedding"],
            index=json["index"],
        )


@dataclass(frozen=True)
class EmbeddingV2Response:
    """
    Response of an OpenAI compatible embedding request

    Parameters:
        object
            The object type, which is always "embedding".
        data
            Embeddings output data.
        model
            Name of the model used to generate the embeddings.
        usage
            Usage information, including the number of tokens in the input and output.
    """

    object: str
    data: List[EmbeddingV2ResponseData]
    model: str
    usage: Usage

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "EmbeddingV2Response":
        return EmbeddingV2Response(
            object=json["object"],
            data=[EmbeddingV2ResponseData.from_json(item) for item in json["data"]],
            model=json["model"],
            usage=Usage.from_json(json["usage"]),
        )



from pathlib import Path
from pytest import raises
from aleph_alpha_client.prompt import Prompt, Image, Text, Tokens
from aleph_alpha_client.prompt_template import PromptTemplate
from liquid.exceptions import LiquidTypeError


def test_to_prompt_with_text_array():
    template = PromptTemplate(
        """
{%- for name in names -%}
Hello {{name}}!
{% endfor -%}
        """
    )
    names = ["World", "Rutger"]

    prompt = template.to_prompt(names=names)

    expected = "".join([f"Hello {name}!\n" for name in names])
    assert prompt == Prompt.from_text(expected)


def test_to_prompt_with_invalid_input():
    template = PromptTemplate(
        """
{%- for name in names -%}
Hello {{name}}!
{% endfor -%}
        """
    )

    with raises(LiquidTypeError):
        template.to_prompt(names=7)


def test_to_prompt_with_single_image(prompt_image: Image):
    template = PromptTemplate(
        """Some Text.
{{whatever}}
More Text
"""
    )

    prompt = template.to_prompt(whatever=template.placeholder(prompt_image))

    expected = Prompt(
        [
            Text.from_text("Some Text.\n"),
            prompt_image,
            Text.from_text("\nMore Text\n"),
        ]
    )
    assert prompt == expected


def test_to_prompt_with_image_sequence(prompt_image: Image):
    template = PromptTemplate(
        """
{%- for image in images -%}
{{image}}
{%- endfor -%}
        """
    )

    prompt = template.to_prompt(
        images=[template.placeholder(prompt_image), template.placeholder(prompt_image)]
    )

    expected = Prompt([prompt_image, prompt_image])
    assert prompt == expected


def test_to_prompt_with_mixed_modality_variables(prompt_image: Image):
    template = PromptTemplate("""{{image}}{{name}}{{image}}""")

    prompt = template.to_prompt(
        image=template.placeholder(prompt_image), name="whatever"
    )

    expected = Prompt([prompt_image, Text.from_text("whatever"), prompt_image])
    assert prompt == expected


def test_to_prompt_with_unused_image(prompt_image: Image):
    template = PromptTemplate("cool")

    prompt = template.to_prompt(images=template.placeholder(prompt_image))

    assert prompt == Prompt.from_text("cool")


def test_to_prompt_with_multiple_different_images(prompt_image: Image):
    image_source_path = Path(__file__).parent / "image_example.jpg"
    second_image = Image.from_file(image_source_path)

    template = PromptTemplate("""{{image_1}}{{image_2}}""")

    prompt = template.to_prompt(
        image_1=template.placeholder(prompt_image),
        image_2=template.placeholder(second_image),
    )

    assert prompt == Prompt([prompt_image, second_image])


def test_to_prompt_with_embedded_prompt(prompt_image: Image):
    user_prompt = Prompt([Text.from_text("Cool"), prompt_image])

    template = PromptTemplate("""{{user_prompt}}""")

    prompt = template.to_prompt(user_prompt=template.embed_prompt(user_prompt))

    assert prompt == user_prompt


def test_to_prompt_does_not_add_whitespace_after_image(prompt_image: Image):
    user_prompt = Prompt([prompt_image, Text.from_text("Cool"), prompt_image])

    template = PromptTemplate("{{user_prompt}}")

    prompt = template.to_prompt(user_prompt=template.embed_prompt(user_prompt))

    assert prompt == user_prompt


def test_to_prompt_skips_empty_strings():
    user_prompt = Prompt(
        [Text.from_text("Cool"), Text.from_text(""), Text.from_text("Also cool")]
    )

    template = PromptTemplate("{{user_prompt}}")

    prompt = template.to_prompt(user_prompt=template.embed_prompt(user_prompt))

    assert prompt == Prompt([Text.from_text("Cool Also cool")])


def test_to_prompt_adds_whitespaces():
    user_prompt = Prompt(
        [Text.from_text("start "), Text.from_text("middle"), Text.from_text(" end")]
    )

    template = PromptTemplate("{{user_prompt}}")

    prompt = template.to_prompt(user_prompt=template.embed_prompt(user_prompt))

    assert prompt == Prompt([Text.from_text("start middle end")])


def test_to_prompt_works_with_tokens():
    user_prompt = Prompt(
        [
            Tokens.from_token_ids([1, 2, 3]),
            Text.from_text("cool"),
            Tokens.from_token_ids([4, 5, 6]),
        ]
    )

    template = PromptTemplate("{{user_prompt}}")

    prompt = template.to_prompt(user_prompt=template.embed_prompt(user_prompt))

    assert prompt == user_prompt


def test_to_prompt_resets_cache(prompt_image: Image):
    user_prompt = Prompt([prompt_image, Text.from_text("Cool"), prompt_image])

    template = PromptTemplate("{{user_prompt}}")

    template.to_prompt(user_prompt=template.embed_prompt(user_prompt))

    assert template.non_text_items == {}



from dataclasses import dataclass, asdict
from typing import (
    Any,
    Dict,
    Mapping,
    Optional,
)
from aleph_alpha_client.prompt import Prompt


@dataclass(frozen=True)
class EvaluationRequest:
    """
    Evaluates the model's likelihood to produce a completion given a prompt.

    Parameters:
        prompt (str, optional, default ""):
            The text to be completed. Unconditional completion can be used with an empty string (default). The prompt may contain a zero shot or few shot task.

        completion_expected (str, required):
            The ground truth completion expected to be produced given the prompt.

        contextual_control_threshold (float, default None)
            If set to None, attention control parameters only apply to those tokens that have
            explicitly been set in the request.
            If set to a non-None value, we apply the control parameters to similar tokens as well.
            Controls that have been applied to one token will then be applied to all other tokens
            that have at least the similarity score defined by this parameter.
            The similarity score is the cosine similarity of token embeddings.

        control_log_additive (bool, default True)
            True: apply control by adding the log(control_factor) to attention scores.
            False: apply control by (attention_scores - - attention_scores.min(-1)) * control_factor

    Examples:
        >>> request = EvaluationRequest(prompt=Prompt.from_text("The api works"), completion_expected=" well")
    """

    prompt: Prompt
    completion_expected: str
    contextual_control_threshold: Optional[float] = None
    control_log_additive: Optional[bool] = True

    def to_json(self) -> Mapping[str, Any]:
        return {**self._asdict(), "prompt": self.prompt.to_json()}

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class EvaluationResponse:
    model_version: str
    message: Optional[str]
    result: Dict[str, Any]
    num_tokens_prompt_total: int

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "EvaluationResponse":
        return EvaluationResponse(
            model_version=json["model_version"],
            result=json["result"],
            num_tokens_prompt_total=json["num_tokens_prompt_total"],
            message=json.get("message"),
        )



import base64
from typing import Any, Dict, Optional, Sequence, Union

from aleph_alpha_client.prompt import Image, Prompt, PromptItem, Text, Tokens, _to_json


class Document:
    """
    A document that can be either a docx document or text/image prompts.
    """

    def __init__(
        self,
        docx: Optional[str] = None,
        prompt: Optional[Sequence[Union[str, Text, Image, Tokens]]] = None,
        text: Optional[str] = None,
    ):
        # We use a base_64 representation for docx documents, because we want to embed the file
        # into a prompt send in JSON.
        self.docx = docx
        self.prompt = prompt
        self.text = text

    @classmethod
    def from_docx_bytes(cls, bytes: bytes):
        """
        Pass a docx file in bytes and prepare it to be used as a document
        """
        docx_base64 = base64.b64encode(bytes).decode()
        return cls(docx=docx_base64)

    @classmethod
    def from_docx_file(cls, path: str):
        """
        Load a docx file from disk and prepare it to be used as a document

        Examples:
            >>> docx_file = "./tests/sample.docx"
            >>> document = Document.from_docx_file(docx_file)
        """
        with open(path, "rb") as f:
            docx_bytes = f.read()
        return cls.from_docx_bytes(docx_bytes)

    @classmethod
    def from_prompt(cls, prompt: Union[Prompt, Sequence[Union[str, Image]]]):
        """
        Pass a prompt that can contain multiple strings and Image prompts and prepare it to be used as a document
        """
        if isinstance(prompt, Prompt):
            return cls(prompt=prompt.items)
        else:
            return cls(prompt=prompt)

    @classmethod
    def from_text(cls, text: str):
        """
        Pass a single text and prepare it to be used as a document

        Example:
            >>> prompt = "This is an example."
            >>> document = Document.from_text(prompt)
        """
        return cls(text=text)

    def _to_serializable_document(self) -> Dict[str, Any]:
        """
        A dict if serialized to JSON is suitable as a document element
        """

        def to_prompt_item(item: Union[str, Image, Text, Tokens]) -> PromptItem:
            # document still uses a plain piece of text for text-prompts
            # -> convert to Text-instance
            return Text.from_text(item) if isinstance(item, str) else item

        if self.docx is not None:
            # Serialize docx to Document JSON format
            return {
                "docx": self.docx,
            }
        elif self.prompt is not None:
            # Serialize prompt to Document JSON format
            prompt_data = [
                _to_json(to_prompt_item(prompt_item)) for prompt_item in self.prompt
            ]
            return {"prompt": prompt_data}
        elif self.text is not None:
            return {
                "text": self.text,
            }
        else:
            raise NotImplementedError("unsupported document type")



from http import HTTPStatus
from random import choice
import re
from typing import Optional
from aleph_alpha_client.aleph_alpha_client import (
    RETRY_STATUS_CODES,
    AsyncClient,
    BusyError,
    Client,
    _raise_for_status,
)
from aleph_alpha_client.completion import CompletionRequest, CompletionResponse
from aleph_alpha_client.prompt import Prompt
import pytest
from pytest_httpserver import HTTPServer
from requests.models import Response


def test_translate_errors():
    response = Response()
    response.code = "bad request"
    response.error_type = "bad request"
    response.status_code = 400
    response._content = b'{ "key" : "a" }'
    with pytest.raises(ValueError):
        _raise_for_status(response.status_code, response.text)


def test_retry_sync(httpserver: HTTPServer):
    num_retries = 2
    client = Client(
        token="TEST_TOKEN", host=httpserver.url_for(""), total_retries=num_retries
    )
    expect_retryable_error(httpserver, num_calls_expected=num_retries)
    expect_valid_version(httpserver)

    client.get_version()


def test_retry_sync_post(httpserver: HTTPServer):
    num_retries = 2
    client = Client(
        host=httpserver.url_for(""), token="TEST_TOKEN", total_retries=num_retries
    )
    expect_retryable_error(httpserver, num_calls_expected=num_retries)
    expect_valid_completion(httpserver)

    request = CompletionRequest(prompt=Prompt.from_text(""), maximum_tokens=7)
    client.complete(request=request, model="model")


def test_exhaust_retries_sync(httpserver: HTTPServer):
    num_retries = 1
    client = Client(
        token="TEST_TOKEN", host=httpserver.url_for(""), total_retries=num_retries
    )
    expect_retryable_error(
        httpserver,
        num_calls_expected=num_retries + 1,
        status_code=HTTPStatus.SERVICE_UNAVAILABLE,
    )
    with pytest.raises(BusyError):
        client.get_version()


async def test_retry_async(httpserver: HTTPServer):
    num_retries = 2
    expect_retryable_error(httpserver, num_calls_expected=num_retries)
    expect_valid_version(httpserver)

    async with AsyncClient(
        token="TEST_TOKEN", host=httpserver.url_for(""), total_retries=num_retries
    ) as client:
        await client.get_version()


async def test_retry_async_post(httpserver: HTTPServer):
    num_retries = 2
    expect_retryable_error(httpserver, num_calls_expected=num_retries)
    expect_valid_completion(httpserver)

    async with AsyncClient(
        token="TEST_TOKEN", host=httpserver.url_for(""), total_retries=num_retries
    ) as client:
        request = CompletionRequest(prompt=Prompt.from_text(""), maximum_tokens=7)
        await client.complete(request, model="FOO")


async def test_exhaust_retries_async(httpserver: HTTPServer):
    num_retries = 1
    expect_retryable_error(
        httpserver,
        num_calls_expected=num_retries + 1,
        status_code=HTTPStatus.SERVICE_UNAVAILABLE,
    )
    with pytest.raises(BusyError):
        async with AsyncClient(
            token="TEST_TOKEN", host=httpserver.url_for(""), total_retries=num_retries
        ) as client:
            await client.get_version()


def expect_retryable_error(
    httpserver: HTTPServer, num_calls_expected: int, status_code: Optional[int] = None
) -> None:
    for i in range(num_calls_expected):
        httpserver.expect_ordered_request(re.compile("^/")).respond_with_data(
            f"error({i})", status=status_code or choice(list(RETRY_STATUS_CODES))
        )


def expect_valid_completion(httpserver: HTTPServer) -> None:
    httpserver.expect_ordered_request("/complete").respond_with_json(
        CompletionResponse(
            model_version="1",
            completions=[],
            num_tokens_prompt_total=0,
            num_tokens_generated=0,
        ).to_json()
    )


def expect_valid_version(httpserver: HTTPServer) -> None:
    httpserver.expect_ordered_request("/version").respond_with_data("ok", status=200)



import base64
import requests


def load_base64_from_url(url: str):
    """
    download a file and return the base64 encoded content
    """
    im_b64 = base64.b64encode(requests.get(url).content).decode()
    return im_b64


def load_base64_from_file(path_and_filename: str):
    """
    load a file from disk and return the base64 encoded content
    """
    with open(path_and_filename, "rb") as f:
        im_b64 = base64.b64encode(f.read()).decode()
    return im_b64
